import re
import pandas as pd
from .agent_tool import *

from .template_text import MRorNot_text, synonyms_text, gwas_id_text, pubmed_text, LLM_MR_template, LLM_MR_MOE_template, \
    LLM_conclusion_template, LLM_Introduction_template, pubmed_text_obo, LLM_template_MR_effect_evaluation
import os

from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image
from reportlab.lib.styles import getSampleStyleSheet
from PyPDF2 import PdfMerger
import time
import sys
from .LLM import llm_chat
import json
import math
from functools import cached_property
from .agent_tool import timer



class MRAgent:
    def __init__(self, mode='O', exposure=None, outcome=None, AI_key=None, model='MR', num=100, bidirectional=False,
                 synonyms=True, introduction=True, LLM_model='gpt-4-turbo-preview', base_url=None, gwas_token=None,
                 opengwas_mode='csv', mr_quality_evaluation=False, mr_quality_evaluation_key_item=None, mrlap=False):
        # 加多一个参数，控制是否从csv中读取gwas列表'csv'或'online'

        self.exposure = exposure
        self.outcome = outcome
        self.AI_key = AI_key
        # 模式选择 OE O E
        self.mode = mode
        # MR模型选择 MR or MR_MOE
        self.model = model
        # 是否双向
        self.bidirectional = bidirectional
        # 是否进行同义词扩充
        self.synonyms = synonyms
        # 是否进行引言
        self.introduction = introduction
        # 若模式为E，则将outcome赋值为exposure，下方步骤中outcome和exposure是等价的
        if self.mode == 'E':
            self.outcome = exposure
        # OE模式已开新类
        # LLM模型选择
        self.LLM_model = LLM_model
        # base_url
        self.base_url = base_url

        # 定义path
        self.define_path()
        # 从pubmed获取文章数量
        self.num = num
        # gwas_token
        self.gwas_token = gwas_token
        # opengwas模式
        self.opengwas_mode = opengwas_mode
        # # if self.opengwas_mode == 'csv':
        #     self.opengwas_csv_init()
        self.mrlap = mrlap

        # MR效果评估
        self.mr_quality_evaluation = mr_quality_evaluation
        if mr_quality_evaluation_key_item == None:
            self.mr_quality_evaluation_key_item = []
        else:
            self.mr_quality_evaluation_key_item = mr_quality_evaluation_key_item

    # 定义path并创建文件夹
    def define_path(self):
        if self.mode == 'OE':
            self.path = os.path.join('output', self.exposure + '_' + self.outcome + '_' + self.LLM_model)
        elif self.mode == 'O':
            self.path = os.path.join('output', self.outcome + '_' + self.LLM_model)
        elif self.mode == 'E':
            self.path = os.path.join('output', self.exposure + '_' + self.LLM_model)

        if not os.path.exists(self.path):
            os.makedirs(self.path)

    # outcome与什么做了相关性但是没有做因果性
    @timer
    def step1(self):
        # 1.1 从pubmed获取相关文章
        pubmed_out = pubmed_crawler(self.outcome, self.num, 'most recent', json_str=False)
        print(pubmed_out)


        # 1.2 输出文章信息
        # 创建一个空列表，用于存放结果
        out = []
        # 提示词
        template = pubmed_text_obo
        for paper in pubmed_out:
            print(paper)
            # 分别取出index, title, abstract
            index = paper['index']
            title = paper['title']
            abstract = paper['abstract']
            # 1.3 LLM判断
            try:
                t = template.format(Outcome=self.outcome, title=title, abstract=abstract)
                gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
                print('gpt_out:')
                print(gpt_out)
                # 利用正则表达式提取结果中的json list
                if '[' in gpt_out and ']' in gpt_out:
                    gpt_out = gpt_out.split('[')[1].split(']')[0]
                    gpt_out = '[' + gpt_out + ']'
                else:
                    gpt_out = '[{"Outcome": null, "Exposure": null}]'
                print('gpt_out_list:')
                print(gpt_out)

                # 1.4 保存结果
                # 转换为json
                items = json.loads(gpt_out)
                print('items:')
                print(items)
                for item in items:
                    # 不为空
                    if item['Outcome'] is not None:
                        # 将index加入item
                        item['index'] = index
                        # 将 title 加入item
                        item['title'] = title
                        # 将 abstract 加入item
                        item['abstract'] = abstract
                        out.append(item)
                print('\n')
            except Exception as e:
                print('error:', e)
        print(out)

        # 1.4 转换为DataFrame
        df = pd.DataFrame(out, columns=['index', 'Outcome', 'Exposure', 'title', 'abstract'])
        # 增加一列nID oeID
        # df['nID'] = range(0, len(df))
        df['oeID'] = range(0, len(df))
        print(df)

        # 1.5 保存
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    # 查看outcome和exposure是否做了MR
    def outcome_exposure_MRorNot(self, outcome, exposure):
        # 休眠1秒
        # global MR_evaluation
        time.sleep(1)
        # 从pubmed获取相关文章
        pub_text = '({Exposure}[Title/Abstract]) AND ({Outcome}[Title/Abstract]) AND (Mendelian randomization[Title/Abstract])'
        pub_text = pub_text.format(Outcome=outcome, Exposure=exposure)
        print(pub_text)
        pubmed_out = pubmed_crawler(pub_text, 10, 'relevance')
        # 调用GPT
        # GPT
        template = MRorNot_text
        t = template.format(Exposure=exposure, Outcome=outcome, pubmed_out=pubmed_out)

        gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
        print(gpt_out)

        # 若输出中包含"Exposure and Outcome were not subjected to Mendelian randomisation."，则说明没有Mendelian randomisation studies
        # 若已经进行了MR则GPT返回的是
        if "Exposure and Outcome were not subjected to Mendelian randomization." in gpt_out:
            print('No')
            mark = 'No'
            strobe_mr_chicklist = 'unavailable'
        else:
            print('Yes')
            mark = 'Yes'
            # mark = gpt_out

            # MR效果评估
            if self.mr_quality_evaluation == False:
                strobe_mr_chicklist = 'unavailable'
            else:
                strobe_mr_chicklist = self.STROBE_MR(gpt_out)
                if strobe_mr_chicklist == None:
                    strobe_mr_chicklist = 'Full text unavailable'
                else:
                    # 按照self.mr_quality_evaluation_key_item list中的关键词进行评估
                    # 以self.mr_quality_evaluation_key_item为key，从strobe_mr_chicklist提取所有的键值对
                    MR_evaluation = {}
                    for key in self.mr_quality_evaluation_key_item:
                        MR_evaluation[key] = strobe_mr_chicklist.get(key, 'unavailable')
                    # print(MR_evaluation)
                    # print(MR_evaluation.values())

                    # 若MR_evaluation存在‘No’，则mark为No
                    if 'no' in MR_evaluation.values():
                        print('no')
                        mark = 'No'
                    else:
                        print('yes')

        # 返回结果mark 和 MR_evaluation
        return mark, gpt_out, strobe_mr_chicklist

    # MR效果评估
    @timer
    def STROBE_MR(self, title):
        paper_details = get_paper_details_pmc(title)
        # 此处获取全文json
        if paper_details == None:
            return None
        else:
            # GPT
            template = LLM_template_MR_effect_evaluation
            t = template.format(paper_details=paper_details)
            gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
            # 正则表达式模式，匹配JSON对象
            json_pattern = r'\{[^}]*\}'

            # 使用正则表达式查找JSON对象
            match = re.search(json_pattern, gpt_out, re.DOTALL)
            # print(match)

            if match:
                json_str = match.group(0)
                # print(json_str)
                try:
                    # 尝试将提取的字符串解析为JSON对象
                    json_obj = json.loads(json_str)
                    print(json_obj)
                except json.JSONDecodeError as e:
                    print("Error decoding JSON:", e)
                    json_obj = None
            else:
                print("No JSON object found in the text.")
                json_obj = None

            return json_obj

    # 第一次查看是否做了MR
    @timer
    def step2(self):
        # 2.1 读取step1的结果
        step1_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df = pd.read_csv(step1_path)
        # print(df)

        # 2.2 查看outcome和exposure是否做了MR
        # 提取outcome和exposure，去重
        df_oe = df[['Exposure', 'Outcome']].drop_duplicates()
        # print(df_oe)
        # 查看df_oe是否做了MR
        df_oe[['MRorNot', 'MR_title', 'STROBE_MR_chicklist']] = df_oe.apply(
            lambda x: self.outcome_exposure_MRorNot(x['Outcome'], x['Exposure']), axis=1).apply(pd.Series)
        # print(df_oe)
        # 以df为基础，将df_oe的MRorNot合并到df中， 'Outcome'和 'Exposure'  是主键
        df = pd.merge(df, df_oe, on=['Exposure', 'Outcome'], how='left')

        # # 2.3 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    #  同义词扩充。提取出所有O和E，新建2表，去重，然后GPT寻找同义词
    @timer
    def step3(self):
        # 3.1 读取step2的结果
        step2_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df = pd.read_csv(step2_path)
        # 去除MRorNot为Yes的行
        df = df[df['MRorNot'] != 'Yes']
        # 判断df是否为空
        if df.empty:
            print('There is no MR studies')
            sys.exit(0)

        # print(df)

        # 3.2 提取所有的Outcome和Exposure
        outcome_list = df['Outcome'].tolist()
        exposure_list = df['Exposure'].tolist()
        # 合并后去重
        all_list = outcome_list + exposure_list
        all_list = list(set(all_list))
        print(all_list)

        # 3.3 建立表2
        df2 = pd.DataFrame(all_list, columns=['OE'])
        # 增加一列nID sID
        # df2['nID'] = range(0, len(df2))
        df2['sID'] = range(0, len(df2))
        # print(df2)

        # 3.4 从df1中提取OE，寻找同义词，并合并到df2中。要求nID继续递增，但是sID还是原来OE的sID
        if self.synonyms:
            for index, row in df2.iterrows():
                # 3.4.1 提取OE和sID
                OE = row['OE']
                sID = row['sID']
                # 3.4.2 调用UMLS
                python_list = get_synonyms(OE, "d6382a8b-5ca8-493a-98ca-2b02fffcaeb5")

                # 3.4.3 合并到df2中
                df3 = pd.DataFrame(python_list, columns=['OE'])
                # 增加一列sID
                df3['sID'] = sID
                print(df3)
                # 合并到df2中
                df2 = pd.concat([df2, df3], ignore_index=True)
                # print(df2)

        # 3.5 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df2.to_csv(out_path, index=False, encoding='utf-8')

    @cached_property
    def opengwas_df(self):
        # 读取opengwas csv文件
        opengwas_path = 'opengwas.csv'
        df = pd.read_csv(opengwas_path)
        df['trait'] = df['trait'].astype(str)
        return df

    @cached_property
    def opengwas_list(self):
        return self.opengwas_df['trait'].tolist()

    def check_keyword_in_opengwas_csv(self, keyword):
        # 字符串匹配,不区分大小写
        return any(keyword.lower() in s.lower() for s in self.opengwas_list)

    @timer
    def step4(self):
        # 4.1 读取step3的结果
        step3_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df = pd.read_csv(step3_path)
        # print(df)

        # 4.2 查看OE是否在opengwas中
        if self.opengwas_mode == 'csv':
            df['opengwas'] = df.apply(lambda x: self.check_keyword_in_opengwas_csv(x['OE']), axis=1)
        elif self.opengwas_mode == 'online':
            df['opengwas'] = df.apply(lambda x: check_keyword_in_opengwas(x['OE']), axis=1)
        print(df)

        # 4.3 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    # 获取OpenGWAS数据库中的GWAS ID
    def get_gwas_id_csv(self, keyword):
        # 根据opengwas_df中的'trait'列，查找与keyword字符串匹配（不区分大小写）的行
        # 查self.opengwas_list找匹配的行，字符串匹配，不区分大小写
        trait_list = []
        for s in self.opengwas_list:
            if keyword.lower() in s.lower():
                trait_list.append(s)
        # trait_list = (s for s in self.opengwas_list if keyword.lower() in s.lower())
        # 如果trait_list长度大于30，则取前30个
        if len(trait_list) > 30:
            trait_list = trait_list[:30]
        # 找到self.opengwas_df中匹配的行
        df = self.opengwas_df[self.opengwas_df['trait'].isin(trait_list)]
        # print(df)

        # 转换为json list
        json_list = []
        # 逐行获取df
        for index, row in df.iterrows():
            # 提取gwas_id
            gwas_id = row['id']
            # 判断year的数据类型为float，则转换为int
            year = row['year']
            if not math.isnan(year):
                year = int(year)

            trait = row['trait']
            consortium = row['consortium']

            samplesize = row['sample_size']
            if not math.isnan(samplesize):
                samplesize = int(samplesize)

            nsnp = row['nsnp']
            if not math.isnan(nsnp):
                nsnp = int(nsnp)

            # # 转换为json 保存了 GWAS ID、Year、Trait、Consortium、Sample size、Number of SNPs
            json_list.append(json.dumps({'gwas_id': gwas_id, 'year': year, 'trait': trait, 'consortium': consortium,
                                         'Sample size': samplesize, 'Number of SNPs': nsnp}))
        print(json_list)

        return json_list

    def step5_get_gwas_id(self, keyword):
        if self.opengwas_mode == 'online':
            json_list = get_gwas_id(keyword)
        elif self.opengwas_mode == 'csv':
            json_list = self.get_gwas_id_csv(keyword)
        # get_gwas_id改为csv模式
        template = gwas_id_text
        t = template.format(keyword=keyword, json_list=json_list)
        gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
        print(gpt_out)
        # 运用正则表达式提取结果中的gwas_id list
        if '[' in gpt_out and ']' in gpt_out:
            gpt_out = gpt_out.split('[')[1].split(']')[0]
            gpt_out = '[' + gpt_out + ']'
        else:
            gpt_out = ['null']

        print(gpt_out)
        # python_list = [item.strip() for item in gpt_out.split(",")]
        # return python_list
        # 取用时再进行处理
        return gpt_out

    # 有了GWAS数据的OE获取所有相关gwas_id
    @timer
    def step5(self):
        # 5.1 读取step4的结果
        step5_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df = pd.read_csv(step5_path)
        # print(df)

        # 5.2 取出opengwas为True的行
        df_oe = df[df['opengwas'] == True]
        # 按照OE去重
        df_oe = df_oe.drop_duplicates(subset=['OE'])

        # 5.3 获取df_oe所有相关gwas_id
        df_oe['gwas_id'] = df_oe.apply(lambda x: self.step5_get_gwas_id(x['OE']), axis=1)

        # 5.4 将df_oe的gwas_id列按OE合并到df中
        df = pd.merge(df, df_oe[['OE', 'gwas_id']], on='OE', how='left')

        # 5.5 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    # 求笛卡尔积，获取所有outcome和exposure可能的组合
    @timer
    def step6(self):
        # 6.1 读取Exposure_and_Outcome.csv
        step1_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df = pd.read_csv(step1_path)
        # 取出MRorNot为No的行
        df_noMR = df[df['MRorNot'] == 'No']
        # print(df_noMR)

        # 6.2 读取Outcome_SNP.csv
        step5_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df_snp = pd.read_csv(step5_path)
        # 取出opengwas为True的行
        # df_snp = df2[df2['opengwas'] == True]

        # 6.3 逐行迭代df_noMR
        for index, row in df_noMR.iterrows():
            # 6.3.1 提取Outcome和Exposure
            Outcome = row['Outcome']
            Exposure = row['Exposure']
            oeID = row['oeID']
            print(Outcome, Exposure)
            # 6.3.2 从df_snp中查找Outcome和Exposure匹配的sID,并分别提取OE为list,并求笛卡尔积
            outcome_id = df_snp[df_snp['OE'] == Outcome]['sID']
            exposure_id = df_snp[df_snp['OE'] == Exposure]['sID']
            # 将提取出的id转换为int
            # print(outcome_id, exposure_id)
            outcome_id = outcome_id.to_numpy()[0]
            exposure_id = exposure_id.to_numpy()[0]
            # print(outcome_id, exposure_id)

            # 6.3.3 按照outcome_id和exposure_id寻找OE，转换为list，求笛卡尔积
            outcome_list = df_snp[df_snp['sID'] == outcome_id]['OE'].to_list()
            exposure_list = df_snp[df_snp['sID'] == exposure_id]['OE'].to_list()
            # print(outcome_list, exposure_list)
            # 求笛卡尔积
            cartesian_product = [(i, j) for i in exposure_list for j in outcome_list]
            # print(cartesian_product)

            # 6.3.4 保存为df_cartesian,合并到df中,oeID为oeID
            df_cartesian = pd.DataFrame(cartesian_product, columns=['Exposure', 'Outcome'])
            df_cartesian['oeID'] = oeID
            # 合并到df中
            df = pd.concat([df, df_cartesian], ignore_index=True)

        # 6.4 查看df中的Outcome和Exposure是否都有snp，匹配df_snp中的OE，查看opengwas列都为True才返回True
        for index, row in df.iterrows():
            # 6.4.1 提取Outcome和Exposure
            Outcome = row['Outcome']
            Exposure = row['Exposure']
            # 6.4.2 查看Outcome和Exposure是否都有snp
            outcome_opengwas = df_snp[df_snp['OE'] == Outcome]['opengwas']
            exposure_opengwas = df_snp[df_snp['OE'] == Exposure]['opengwas']
            # print(outcome_opengwas, exposure_opengwas)
            # 6.4.3 查看opengwas列都为True才返回True
            if outcome_opengwas.all() and exposure_opengwas.all():
                df.loc[index, 'opengwas'] = True
            else:
                df.loc[index, 'opengwas'] = False

        # 6.4 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    # Exposure_and_Outcome中，opengwas为TRUE的，并MRorNot这里为Nan的，查看是否做了MR
    @timer
    def step7(self):
        # 7.1 读取Exposure_and_Outcome.csv
        step1_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df = pd.read_csv(step1_path)
        # 按Outcome和Exposure去重，保留第一次出现的
        df2 = df.drop_duplicates(subset=['Exposure', 'Outcome'])
        print(df2)
        # 取出opengwas为True的行
        df_mr = df2[df2['opengwas'] == True]
        print(df_mr)
        # 取出MRorNot为Nan的行
        df_mr = df_mr[df_mr['MRorNot'].isnull()]
        print(df_mr)
        # df_mr去重
        df_mr = df_mr.drop_duplicates(subset=['Exposure', 'Outcome'])
        print(df_mr)

        # 判断df_mr不为空
        if not df_mr.empty:
            # 7.2 查看outcome和exposure是否做了MR
            df_mr[['MRorNot', 'MR_title', 'STROBE_MR_chicklist']] = df_mr.apply(
                lambda x: self.outcome_exposure_MRorNot(x['Outcome'], x['Exposure']), axis=1).apply(pd.Series)

            # 按Outcome和Exposure合并到df中，'Outcome'和 'Exposure'  是主键，替换MRorNot列
            df = pd.merge(df, df_mr[['Exposure', 'Outcome', 'MRorNot']], on=['Exposure', 'Outcome'], how='left',
                          suffixes=('', '_y'))
            # 使用df_mr中的'MRorNot'列替换df中的NaN值
            df['MRorNot'] = df['MRorNot'].combine_first(df['MRorNot_y'])
            # 删除多余的'MRorNot_y'列
            df = df.drop(columns='MRorNot_y')

        # 7.3 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df.to_csv(out_path, index=False, encoding='utf-8')

    # 最终选出Outcome和Exposure进行MR
    @timer
    def step8(self):
        # 8.1 读取Exposure_and_Outcome.csv
        step1_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df = pd.read_csv(step1_path)

        # 8.2 选出所有Outcome和Exposure都没有进行MR的oeID
        # 创建新的df空表df_mr
        df_mr = pd.DataFrame(columns=df.columns)

        # 提取所有的oeID
        oeID_set = set(df['oeID'].tolist())
        print(oeID_set)
        for i in oeID_set:
            # print(i)
            df_oeID = df[df['oeID'] == i]
            # print(df_oeID)
            # 查看是否有MRorNot为Yes的行，若没有，则选出来放入df_mr
            if not 'Yes' in df_oeID['MRorNot'].tolist():
                # print(df_oeID)
                df_mr = pd.concat([df_mr, df_oeID], ignore_index=True)

        # df_mr去重 去除opengwas为False的行
        # print(df_mr)
        df_mr = df_mr.drop_duplicates(subset=['Exposure', 'Outcome'])
        df_mr = df_mr[df_mr['opengwas'] == True]
        print(df_mr)

        # 8.3 保存为csv
        # 将结果写入CSV文件,utf-8编码
        out_path = os.path.join(self.path, 'mr_run.csv')
        df_mr.to_csv(out_path, index=False, encoding='utf-8')

    # 调用GPT解释MR的结果
    @timer
    def LLM_MR_result(self, Exposure, Outcome, Exposure_id, Outcome_id, snp_path):
        print(Exposure, Outcome, Exposure_id, Outcome_id)

        if self.model == 'MR':
            # 打开文件
            with open(os.path.join(snp_path, 'table.MRresult.csv'), 'r') as file:
                MRresult = file.read()
            with open(os.path.join(snp_path, 'table.heterogeneity.csv'), 'r') as file:
                heterogeneity = file.read()
            with open(os.path.join(snp_path, 'table.pleiotropy.csv'), 'r') as file:
                pleiotropy = file.read()

            template = LLM_MR_template
            t = template.format(Outcome=Outcome, Exposure=Exposure, MRresult=MRresult, heterogeneity=heterogeneity,
                                pleiotropy=pleiotropy, Exposure_id=Exposure_id, Outcome_id=Outcome_id)
            gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
            # print(gpt_out)
            # 保存输出结果
            with open(os.path.join(snp_path, 'LLM_result.txt'), 'w') as file:
                file.write(gpt_out)
        elif self.model == 'MR_MOE':
            # 打开文件
            with open(os.path.join(snp_path, 'MR.MRresult.csv'), 'r') as file:
                MRresult = file.read()
            with open(os.path.join(snp_path, 'MR.heterogeneity.csv'), 'r') as file:
                heterogeneity = file.read()
            with open(os.path.join(snp_path, 'MR.table.pleiotropy.csv'), 'r') as file:
                pleiotropy = file.read()

            template = LLM_MR_MOE_template
            t = template.format(Outcome=Outcome, Exposure=Exposure, MRresult=MRresult, heterogeneity=heterogeneity,
                                pleiotropy=pleiotropy, Exposure_id=Exposure_id, Outcome_id=Outcome_id)
            gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
            # 保存输出结果
            with open(os.path.join(snp_path, 'LLM_result.txt'), 'w') as file:
                file.write(gpt_out)
        else:
            gpt_out = None

        if self.mrlap:
            mrlap_text = self.mrlap_result_text(snp_path)
        else:
            mrlap_text = None

        self.LLM_MR_result_PDF(Exposure, Outcome, gpt_out, mrlap_text, snp_path)

        return gpt_out

    def LLM_MR_result_PDF(self, Exposure, Outcome, gpt_out, mrlap_text, snp_path):
        # 创建PDF####################################
        doc = SimpleDocTemplate(os.path.join(snp_path, "Report.pdf"), pagesize=letter)
        # 设置样式
        styles = getSampleStyleSheet()
        # 创建一个空的 Story 列表来保存文档内容
        story = []
        # 添加标题
        title = Paragraph("A Mendelian randomisation study about " + Exposure + " and " + Outcome,
                          styles["Title"])
        story.append(title)
        story.append(Spacer(1, 12))
        styles["BodyText"].fontSize = 12  # 设置字体大小为 14
        styles["BodyText"].alignment = 4  # 设置文字居中 4
        styles["BodyText"].fontName = 'Times-Roman'  # 设置字体
        # 分析结果
        subtitle = Paragraph("Analysis of MR results", styles["Heading2"])
        story.append(subtitle)
        gpt_out_list = gpt_out.split('\n')
        for gpt_out_i in gpt_out_list:
            text = Paragraph(gpt_out_i, styles["BodyText"])
            # story.append(Spacer(1, 12))
            story.append(text)
        # MRlap结果
        if self.mrlap:
            subtitle = Paragraph("Analysis of sample overlap", styles["Heading2"])
            story.append(subtitle)
            mrlap_text_list = mrlap_text.split('\n')
            for mrlap_text_i in mrlap_text_list:
                text = Paragraph(mrlap_text_i, styles["BodyText"])
                story.append(text)
        # 构建 PDF
        doc.build(story)
        # 合并PDF和图片
        # 创建一个 PdfFileMerger 对象
        merger = PdfMerger()
        # 列出你想要合并的 PDF 文件
        pdfs = [os.path.join(snp_path, "Report.pdf"), os.path.join(snp_path, 'pic.scatter_plot.pdf'),
                os.path.join(snp_path, 'pic.forest.pdf'), os.path.join(snp_path, 'pic.funnel_plot.pdf'),
                os.path.join(snp_path, 'pic.leaveoneout.pdf')]
        # 循环遍历 PDF 文件并将它们添加到合并器
        for pdf in pdfs:
            merger.append(pdf)
        # 写入合并的 PDF 文件
        merger.write(os.path.join(snp_path, "Report.pdf"))
        # 关闭合并器
        merger.close()

    @timer
    def mrlap_result_text(self, snp_path):
        # 读取json文件
        with open(os.path.join(snp_path, 'MRlap_results.json'), 'r') as file:
            mrlap_json = json.load(file)
        print(mrlap_json)
        # 提取"MRcorrection"中的  "corrected_effect" "corrected_effect_se" "corrected_effect_p" "test_difference" "p_difference"
        MRcorrection = mrlap_json['MRcorrection']
        corrected_effect = MRcorrection['corrected_effect'][0]
        corrected_effect_se = MRcorrection['corrected_effect_se'][0]
        corrected_effect_p = MRcorrection['corrected_effect_p'][0]
        test_difference = MRcorrection['test_difference'][0]
        p_difference = MRcorrection['p_difference'][0]
        mrlap_text = f"""
    In Mendelian Randomization (MR) analyses, sample overlap between the exposure and outcome datasets can introduce bias, potentially leading to inaccurate causal effect estimates. MRlap addresses this issue by using cross-trait Linkage Disequilibrium Score Regression (LDSC) to correct for biases arising from sample overlap, weak instruments, and winner’s curse. This correction provides a more robust estimate of the causal effect.
    
    In our analysis using MRlap, the corrected causal effect estimate ("corrected_effect") was {corrected_effect}, with a standard error ("corrected_effect_se") of {corrected_effect_se}. The p-value for the corrected causal effect ("corrected_effect_p") was {corrected_effect_p}. Given that the p-value is {'<0.05' if corrected_effect_p < 0.05 else '>0.05'}, the corrected effect is {'statistically significant' if corrected_effect_p < 0.05 else 'not statistically significant'}.
    
    Furthermore, the test statistic used to compare the observed and corrected effects ("test_difference") was {test_difference}, with a corresponding p-value ("p_difference") of {p_difference}. Since the p-value is {'<0.05' if p_difference < 0.05 else '>0.05'}, this indicates that {'the correction significantly impacted the effect estimate' if p_difference < 0.05 else 'the correction did not significantly impact the effect estimate'}.
    
    """
        # 追加字段到文件LLM_result.txt中
        with open(os.path.join(snp_path, 'LLM_result.txt'), 'a') as file:
            file.write(mrlap_text)
        return mrlap_text

    def LLM_Introduction(self, Exposure, Outcome, path):
        # GPT写引言####################################
        # pubmed获取Outcome相关文章
        pub_text = '{Outcome}[Title/Abstract]'
        pub_text = pub_text.format(Outcome=Outcome)
        print(pub_text)
        Outcome_pubmed = pubmed_crawler(pub_text, 50, 'relevance')
        # pubmed获取Exposure相关文章
        pub_text = '{Exposure}[Title/Abstract]'
        pub_text = pub_text.format(Exposure=Exposure)
        print(pub_text)
        Exposure_pubmed = pubmed_crawler(pub_text, 50, 'relevance')
        # pubmed获取Exposure和Outcome相关文章
        pub_text = '(({Exposure}[Title/Abstract]) AND ({Outcome}[Title/Abstract]))'
        pub_text = pub_text.format(Exposure=Exposure, Outcome=Outcome)
        print(pub_text)
        Exposure_Outcome_pubmed = pubmed_crawler(pub_text, 50, 'relevance')
        # GPT
        template = LLM_Introduction_template
        t = template.format(Outcome=Outcome, Exposure=Exposure, Outcome_pubmed=Outcome_pubmed,
                            Exposure_pubmed=Exposure_pubmed, Exposure_Outcome_pubmed=Exposure_Outcome_pubmed)
        gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
        # 保存输出结果
        with open(os.path.join(path, 'LLM_Introduction.txt'), 'w', encoding='utf-8') as file:
            file.write(gpt_out)

        ### 获取PubMed数据库中参考文章的详细信息####################################
        # O或E时才获取文章详细信息
        if self.mode == 'O' or self.mode == 'E':
            # 读取CSV文件
            df = pd.read_csv(os.path.join(self.path, 'Exposure_and_Outcome.csv'))
            df = df[(df['Exposure'] == Exposure) & (df['Outcome'] == Outcome)]
            print(df)
            df = df[['title']]
            print(df)
            # 去重
            df = df.drop_duplicates()
            # 转为list
            title_list = df['title'].tolist()
            # 获取文章详细信息
            details_list = []
            for title in title_list:
                # print(title)
                details_list.append(get_paper_details(title))

        self.LLM_Introduction_PDF(Exposure, Outcome, details_list, gpt_out, path)

    def LLM_Introduction_PDF(self, Exposure, Outcome, details_list, gpt_out, path):
        # 创建PDF####################################
        doc = SimpleDocTemplate(os.path.join(path, "Introduction.pdf"), pagesize=letter)
        # 设置样式
        styles = getSampleStyleSheet()
        # 创建一个空的 Story 列表来保存文档内容
        story = []
        # 添加标题
        title = Paragraph("A Mendelian randomisation study about " + Exposure + " and " + Outcome, styles["Title"])
        story.append(title)
        story.append(Spacer(1, 12))
        styles["BodyText"].fontSize = 12  # 设置字体大小为 14
        styles["BodyText"].alignment = 4  # 设置文字居中 4
        styles["BodyText"].fontName = 'Times-Roman'  # 设置字体
        # 引言
        subtitle = Paragraph("Introduction", styles["Heading2"])
        story.append(subtitle)
        # 引言内容
        gpt_out_list = gpt_out.split('\n')
        for gpt_out_i in gpt_out_list:
            text = Paragraph(gpt_out_i, styles["BodyText"])
            story.append(text)
        if self.mode == 'O' or self.mode == 'E':
            # 参考文献
            subtitle = Paragraph("Important References", styles["Heading2"])
            story.append(subtitle)
            for details in details_list:
                text = Paragraph('Title: ' + details[0], styles["BodyText"])
                story.append(text)
                text = Paragraph('PMID: ' + details[1], styles["BodyText"])
                story.append(text)
                text = Paragraph('DOI: ' + details[2], styles["BodyText"])
                story.append(text)
                text = Paragraph('Journal:' + details[3], styles["BodyText"])
                story.append(text)
                text = Paragraph('Author: ' + details[4], styles["BodyText"])
                story.append(text)
                text = Paragraph('Year: ' + details[5], styles["BodyText"])
                story.append(text)
                text = Paragraph('Abstract: ' + details[6], styles["BodyText"])
                story.append(text)
                story.append(Spacer(1, 12))
        # 构建 PDF
        doc.build(story)

    # 总结各个SNP的结果，生成一个综合结果和PDF
    def LLM_conclusion(self, Exposure, Outcome, path):
        # 调用GPT解释MR的结果
        # 打开文件
        with open(os.path.join(path, self.model + '_LLM_result_all.txt'), 'r') as file:
            LLM_result_all = file.read()
        template = LLM_conclusion_template
        t = template.format(Outcome=Outcome, Exposure=Exposure, MRresult=LLM_result_all)
        gpt_out = llm_chat(t, self.LLM_model, self.AI_key, self.base_url)
        # 保存输出结果
        with open(os.path.join(path, self.model + '_LLM_result_all_conclusion.txt'), 'w', encoding='utf-8') as file:
            file.write(gpt_out)

        # # 生成PDF####################################
        # 创建
        doc = SimpleDocTemplate(os.path.join(path, "Conclusion.pdf"), pagesize=letter)
        # 设置样式
        styles = getSampleStyleSheet()
        # 创建一个空的 Story 列表来保存文档内容
        story = []
        # 添加标题
        title = Paragraph("A Mendelian randomisation study about " + Exposure + " and " + Outcome,
                          styles["Title"])
        story.append(title)
        story.append(Spacer(1, 12))

        # 添加Conclusion
        subtitle = Paragraph("Conclusion", styles["Heading2"])
        story.append(subtitle)

        gpt_out_list = gpt_out.split('\n')
        for gpt_out_i in gpt_out_list:
            text = Paragraph(gpt_out_i, styles["BodyText"])
            # story.append(Spacer(1, 12))
            story.append(text)

        # 构建 PDF
        doc.build(story)

    # MRLap
    @timer
    def step9_mrlap(self, Exposure_id, Outcome_id, path):
        # 从self.opengwas_df中查找Exposure_id和Outcome_id的sample_size
        Exposure_sample_size = self.opengwas_df[self.opengwas_df['id'] == Exposure_id]['sample_size'].values[0]
        Outcome_sample_size = self.opengwas_df[self.opengwas_df['id'] == Outcome_id]['sample_size'].values[0]
        # 运行
        MRtool_MRlap(Exposure_id, Outcome_id, path, Exposure_sample_size, Outcome_sample_size)

    def step9_run_mr_LLM(self, Exposure, Outcome, path, cartesian_product):
        for i, j in cartesian_product:
            # i = str(i)
            # j = str(j)
            try:
                print(i, j)
                if self.model == 'MR':
                    snp_path = os.path.join(path, 'MR_' + i + '_' + j)
                    if not os.path.exists(snp_path):
                        os.makedirs(snp_path)
                    snp_path = snp_path.replace('\\', '//')
                    print(snp_path)
                    MRtool(i, j, snp_path, self.gwas_token)
                    if self.mrlap:
                        self.step9_mrlap(i, j, snp_path)
                    # 调用GPT解释MR的结果
                    self.LLM_MR_result(Exposure=Exposure, Outcome=Outcome, Exposure_id=i, Outcome_id=j,
                                       snp_path=snp_path)
                elif self.model == 'MR_MOE':
                    snp_path = os.path.join(path, 'MR_MOE_' + i + '_' + j)
                    if not os.path.exists(snp_path):
                        os.makedirs(snp_path)
                    snp_path = snp_path.replace('\\', '//')
                    print(snp_path)
                    MRtool_MOE(i, j, snp_path, self.gwas_token)
                    if self.mrlap:
                        self.step9_mrlap(i, j, snp_path)
                    self.LLM_MR_result(Exposure=Exposure, Outcome=Outcome, Exposure_id=i, Outcome_id=j,
                                       snp_path=snp_path)

            except Exception as e:
                print('Step9_run_mr_LLM Error')
                print(e)

        # 总结各个SNP的结果，生成一个综合结果和PDF
        # 读取snp_path下的所有LLM_result.txt
        # 合并为一个文件
        print(Exposure, Outcome, 'conclusion')
        k = 0
        for i, j in cartesian_product:
            try:
                if self.model == 'MR':
                    snp_path = os.path.join(path, 'MR_' + i + '_' + j, 'LLM_result.txt')
                elif self.model == 'MR_MOE':
                    snp_path = os.path.join(path, 'MR_MOE_' + i + '_' + j, 'LLM_result.txt')
                with open(snp_path, 'r') as file:
                    LLM_result = file.read()
                with open(os.path.join(path, self.model + '_LLM_result_all.txt'), 'a') as file:
                    # 写入gwas_id
                    file.write('index: ' + str(k) + '\n')
                    file.write('Exposure_id: ' + i + '\n')
                    file.write('Outcome_id: ' + j + '\n')
                    # 写入文件
                    file.write(LLM_result)
                    file.write('\n\n')
                    k += 1
            except Exception as e:
                print(e)
                print('Error')
        # 生成PDF
        try:
            self.LLM_conclusion(Exposure, Outcome, path)
        except Exception as e:
            print(e)
            print('LLM_conclusion Error')

    # 确定cartesian_product的poppulation是否一致
    @timer
    def step9_gwas_poppulation(self, cartesian_product):
        out_cartesian_product = []
        for i, j in cartesian_product:
            # 从self.opengwas_df中查找i和j的population
            i_population = self.opengwas_df[self.opengwas_df['id'] == i]['population'].values[0]
            j_population = self.opengwas_df[self.opengwas_df['id'] == j]['population'].values[0]
            if i_population == j_population:
                out_cartesian_product.append((i, j))
        return out_cartesian_product

    # 运行MR
    @timer
    def step9(self):
        # 9.1 读取mr_run.csv
        global snp_path
        step9_path = os.path.join(self.path, 'mr_run.csv')
        print(step9_path)
        df = pd.read_csv(step9_path)
        # print(df)
        # 读取Outcome_snp.csv
        step5_path = os.path.join(self.path, 'Outcome_SNP.csv')
        df_snp = pd.read_csv(step5_path)
        # 读取Exposure_and_Outcome.csv
        step1_path = os.path.join(self.path, 'Exposure_and_Outcome.csv')
        df_oe = pd.read_csv(step1_path)

        # 9.2 按oeID运行MR
        # 提取所有的oeID
        oeID_set = set(df['oeID'].tolist())
        print(oeID_set)
        for id in oeID_set:
            df_oeID = df[df['oeID'] == id]
            # print(df_oeID)
            # 从df_oe中提取oeID匹配的第一个Outcome和Exposure
            oe_Outcome = df_oe[df_oe['oeID'] == id]['Outcome'].to_list()[0]
            oe_Exposure = df_oe[df_oe['oeID'] == id]['Exposure'].to_list()[0]
            print(oe_Exposure, oe_Outcome)

            # 创建文件夹
            oe_path = os.path.join(self.path, oe_Exposure + '_' + oe_Outcome)
            if not os.path.exists(oe_path):
                os.makedirs(oe_path)

            # 写introduction
            if self.introduction:
                self.LLM_Introduction(Exposure=oe_Exposure, Outcome=oe_Outcome, path=oe_path)

            # 循环oeID 取同义词
            for index, row in df_oeID.iterrows():
                Outcome = row['Outcome']
                Exposure = row['Exposure']
                oeID = row['oeID']
                print(Exposure, Outcome, oeID)
                # 在df_snp中查找Outcome和Exposure匹配的gwad_id
                Outcome_id = df_snp[df_snp['OE'] == Outcome]['gwas_id']
                Exposure_id = df_snp[df_snp['OE'] == Exposure]['gwas_id']

                # 将提取出的python list形式的str转换为list
                Outcome_id = Outcome_id.to_numpy()[0]
                Exposure_id = Exposure_id.to_numpy()[0]
                print(Outcome_id, Exposure_id)
                Outcome_id_list = eval(Outcome_id)
                Exposure_id_list = eval(Exposure_id)
                print(Outcome_id_list, Exposure_id_list)

                # 创建文件夹
                path = os.path.join(oe_path, Exposure + '_' + Outcome)
                if not os.path.exists(path):
                    os.makedirs(path)
                # Outcome_id_list, Exposure_id_list 笛卡尔积
                # 此处应该加入人群种族的判断
                cartesian_product = [(i, j) for i in Exposure_id_list for j in Outcome_id_list]
                if self.opengwas_mode == 'csv':
                    cartesian_product = self.step9_gwas_poppulation(cartesian_product)
                print(cartesian_product)
                # 运行MR
                self.step9_run_mr_LLM(Exposure=Exposure, Outcome=Outcome, path=path,
                                      cartesian_product=cartesian_product)

                # 双向
                if self.bidirectional:
                    path = os.path.join(oe_path, Outcome + '_' + Exposure)
                    if not os.path.exists(path):
                        os.makedirs(path)
                    cartesian_product = [(j, i) for i in Exposure_id_list for j in Outcome_id_list]
                    if self.opengwas_mode == 'csv':
                        cartesian_product = self.step9_gwas_poppulation(cartesian_product)
                    print(cartesian_product)
                    # 运行MR
                    self.step9_run_mr_LLM(Exposure=Outcome, Outcome=Exposure, path=path,
                                          cartesian_product=cartesian_product)

    def run(self, step=None):
        # 判断是否需要进行step1
        if step is None:
            step = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        if 1 in step:
            print('#########step1#########')
            self.step1()
        if 2 in step:
            print('#########step2#########')
            self.step2()
        if 3 in step:
            print('#########step3#########')
            self.step3()
        if 4 in step:
            print('#########step4#########')
            self.step4()
        if 5 in step:
            print('#########step5#########')
            self.step5()
        if 6 in step:
            print('#########step6#########')
            self.step6()
        if 7 in step:
            print('#########step7#########')
            self.step7()
        if 8 in step:
            print('#########step8#########')
            self.step8()
        if 9 in step:
            print('#########step9#########')
            self.step9()






from Bio import Entrez
import requests
from bs4 import BeautifulSoup

import os
import json

import time
import urllib

def timer(func):
    def func_wrapper(*args, **kwargs):
        from time import time
        time_start = time()
        result = func(*args, **kwargs)
        time_end = time()
        time_spend = time_end - time_start
        print('%s cost time: %.3f s' % (func.__name__, time_spend))
        return result

    return func_wrapper



# 爬虫爬取PubMed数据
def pubmed_crawler(keyword, num_records, sort_order, json_str=True):
    Entrez.email = '670525744@qq.com'  # 请替换为你的邮箱

    def search_pubmed(keyword, num_records, sort_order):
        handle = Entrez.esearch(db='pubmed',
                                sort=sort_order,
                                retmax=str(num_records),
                                retmode='xml',
                                term=keyword)
        results = Entrez.read(handle)
        return results

    def fetch_details(id_list):
        ids = ','.join(id_list)
        handle = Entrez.efetch(db='pubmed',
                               retmode='xml',
                               id=ids)
        results = Entrez.read(handle, validate=False)
        return results

    def get_paper_details(paper):
        try:
            title = paper['MedlineCitation']['Article']['ArticleTitle']
        except KeyError:
            title = 'NULL'

        try:
            abstract = paper['MedlineCitation']['Article']['Abstract']['AbstractText'][0]
        except KeyError:
            abstract = 'NULL'

        return title, abstract

    def search_and_print_papers(keyword, num_records, sort_order):
        results = search_pubmed(keyword, num_records, sort_order)
        # print(results)
        id_list = results['IdList']
        if len(id_list) == 0:
            return json.dumps([{
                'index': 0,
                'title': 'No paper found',
                'abstract': 'No paper found'
            }])
        papers = fetch_details(id_list)
        # print(papers)
        paper_details = []
        for i, paper in enumerate(papers['PubmedArticle'], start=1):
            title, abstract = get_paper_details(paper)
            paper_details.append({
                'index': i,
                'title': title,
                'abstract': abstract
            })
        if json_str:
            return json.dumps(paper_details)
        else:
            return paper_details

    return search_and_print_papers(keyword, num_records, sort_order)


# 获取PubMed数据库中某个文章的详细信息
def get_paper_details(paper_title):
    Entrez.email = 'xuwei_chn@foxmail.com'  # 请替换为你的邮箱

    def search_pubmed(keyword, num_records, sort_order):
        handle = Entrez.esearch(db='pubmed',
                                sort=sort_order,
                                retmax=str(num_records),
                                retmode='xml',
                                term=keyword)
        results = Entrez.read(handle)
        return results

    def fetch_details(id_list):
        ids = ','.join(id_list)
        handle = Entrez.efetch(db='pubmed',
                               retmode='xml',
                               id=ids)
        results = Entrez.read(handle)
        return results

    def search_and_print_papers(keyword):
        results = search_pubmed(keyword, 1, 'relevance')
        # print(results)
        id_list = results['IdList']
        papers = fetch_details(id_list)
        print(papers)
        paper = papers['PubmedArticle'][0]

        try:
            title = paper['MedlineCitation']['Article']['ArticleTitle']
        except KeyError:
            title = 'NULL'

        try:
            abstract = paper['MedlineCitation']['Article']['Abstract']['AbstractText'][0]
        except KeyError:
            abstract = 'NULL'

        try:
            PMID = paper['MedlineCitation']['PMID']
        except KeyError:
            PMID = 'NULL'

        try:
            DOI = paper['MedlineCitation']['Article']['ELocationID'][0]
        except KeyError:
            DOI = 'NULL'

        try:
            Journal = paper['MedlineCitation']['Article']['Journal']['Title']
        except KeyError:
            Journal = 'NULL'

        try:
            Author = paper['MedlineCitation']['Article']['AuthorList'][0]['LastName']
        except KeyError:
            Author = 'NULL'

        try:
            Year = paper['MedlineCitation']['Article']['Journal']['JournalIssue']['PubDate']['Year']
        except KeyError:
            Year = 'NULL'

        print(title)
        print(PMID)
        print(DOI)
        print(Journal)
        print(Author)
        print(Year)
        print(abstract)
        return [title, PMID, DOI, Journal, Author, Year, abstract]

    return search_and_print_papers(paper_title)


# 检测关键词是否在OpenGWAS数据库中有对应的SNP
def check_keyword_in_opengwas(keyword):
    # OpenGWAS网页地址
    url = f"https://gwas.mrcieu.ac.uk/datasets/?trait__icontains={keyword}"

    # 发送GET请求
    response = requests.get(url)

    # 检查响应状态码
    if response.status_code == 200:
        # 如果状态码为200，表示请求成功
        soup = BeautifulSoup(response.text, 'html.parser')
        # print(soup)
        # print(soup.text)
        # 检查是否存在指定的文本
        if "Filtered to 0 records" in soup.text:
            print(f"No SNP related to '{keyword}' was found in the OpenGWAS database.")
            return False
        else:
            print(f"SNP(s) related to '{keyword}' exist in the OpenGWAS database.")
            return True
    else:
        # 其他状态码表示发生了错误
        print(f"An error occurred while querying the OpenGWAS database. Status code: {response.status_code}")
        return None


# 获取OpenGWAS数据库中的GWAS ID
def get_gwas_id(keyword):
    def remove_newlines(lst):
        # Remove /n from each string in a list
        return [s.replace('\n', '') for s in lst]

    data = []
    for i in range(1, 11):
        # OpenGWAS网页地址
        url = f"https://gwas.mrcieu.ac.uk/datasets/?trait__icontains={keyword}&page={i}"
        response = requests.get(url)
        soup = BeautifulSoup(response.text, 'html.parser')
        table = soup.find('table')
        headers = [header.text for header in table.find_all('th')]
        headers = remove_newlines(headers)
        rows = table.find_all('tr')
        for row in rows[1:]:
            values = [value.text for value in row.find_all('td')]
            data.append(dict(zip(headers, values)))

    # data去重
    data = [dict(t) for t in {tuple(d.items()) for d in data}]
    print(data)

    return data


def get_paper_details_pmc(paper_title):
    Entrez.email = 'xuwei_chn@foxmail.com'  # 请替换为你的邮箱

    def search_pubmed(keyword, num_records, sort_order):
        handle = Entrez.esearch(db='pubmed',
                                sort=sort_order,
                                retmax=str(num_records),
                                retmode='xml',
                                term=keyword)
        results = Entrez.read(handle)
        return results

    def get_pcm_full_text(pmc_id):
        # 直接访问接口网站获取json全文
        # https://www.ncbi.nlm.nih.gov/research/bionlp/RESTful/pmcoa.cgi/BioC_json/[ID]/unicode
        url = f"https://www.ncbi.nlm.nih.gov/research/bionlp/RESTful/pmcoa.cgi/BioC_json/{pmc_id}/unicode"
        print(url)
        response = urllib.request.urlopen(url)
        full_text = response.read().decode('utf-8')
        return full_text

    def search_and_print_papers(keyword):
        results = search_pubmed(keyword, 1, 'relevance')
        id_list = results['IdList']
        if len(id_list) == 0:
            print("No papers found for the given query.")
            return None
        else:
            paper_id = id_list[0]
            full_text = get_pcm_full_text(paper_id)
            if 'No record can be found for the input' in full_text:
                print("No full text found for the given query.")
                return None
            else:
                return full_text

    return search_and_print_papers(paper_title)

@timer
def MRtool(Exposure_id, Outcome_id, path, gwas_token):
    # 等待5s
    time.sleep(5)

    r_script = """
    #引用包
    Sys.setenv(OPENGWAS_JWT="{gwas_token}")
    library(TwoSampleMR)
    library(ieugwasr)
    
    # options(ieugwasr_api = 'gwas-api.mrcieu.ac.uk/')
    
    
    num_rows <- 0
    tryCatch({{
      p_value <- 5e-08
      exposure_dat <- extract_instruments(outcomes = '{Exposure_id}',
                                          p1=p_value,
                                          clump=TRUE,
                                          r2=0.001,
                                          kb=5000
      )
      num_rows <- nrow(exposure_dat)
      print(num_rows)
    }}, error = function(e) {{
      message("First time gwas data error.", e$message)
    }})
    
    if (num_rows < 11) {{
      tryCatch({{
        p_value <- 5e-06
        exposure_dat <- extract_instruments(outcomes = '{Exposure_id}',
                                            p1=p_value,
                                            clump=TRUE,
                                            r2=0.001,
                                            kb=5000
        )
        num_rows <- nrow(exposure_dat)
        print(num_rows)
      }}, error = function(e) {{
        message("Second time gwas data error. ", e$message)
      }})
    }}
    
    if (num_rows < 11) {{
      tryCatch({{
        p_value <- 5e-05
        exposure_dat <- extract_instruments(outcomes = '{Exposure_id}',
                                            p1=p_value,
                                            clump=TRUE,
                                            r2=0.001,
                                            kb=5000
        )
        num_rows <- nrow(exposure_dat)
        print(num_rows)
      }}, error = function(e) {{
        message("Third time gwas data error.", e$message)
      }})
    }}
    
    
    outcomeID="{Outcome_id}"
    #提取结局数据
    outcome_dat <- extract_outcome_data(snps=exposure_dat$SNP, outcomes=outcomeID)
    dat <- harmonise_data(exposure_dat, outcome_dat)
    outTab=dat[dat$mr_keep=="TRUE",]
    write.csv(outTab, file=".//{path}//table.SNP.csv", row.names=F)

    #孟德尔随机化分析
    mrResult=mr(dat)
    #mr_method_list()$obj
    # mrResult=mr(dat, method_list=c("mr_ivw", "mr_egger_regression", "mr_weighted_median", "mr_simple_mode", "mr_weighted_mode"))
    #对结果进行OR值的计算
    mrTab=generate_odds_ratios(mrResult)
    write.csv(mrTab, file=".//{path}//table.MRresult.csv", row.names=F)


    #异质性分析
    heterTab=mr_heterogeneity(dat)
    write.csv(heterTab, file=".//{path}//table.heterogeneity.csv", row.names=F)

    #多效性检验
    pleioTab=mr_pleiotropy_test(dat)
    write.csv(pleioTab, file=".//{path}//table.pleiotropy.csv", row.names=F)

    #绘制散点图
    pdf(file=".//{path}//pic.scatter_plot.pdf", width=7.5, height=7)
    mr_scatter_plot(mrResult, dat)
    dev.off()

    #森林图
    res_single=mr_singlesnp(dat)      #得到每个工具变量对结局的影响
    pdf(file=".//{path}//pic.forest.pdf", width=7, height=5.5)
    mr_forest_plot(res_single)
    dev.off()

    #漏斗图
    pdf(file=".//{path}//pic.funnel_plot.pdf", width=7, height=6.5)
    mr_funnel_plot(singlesnp_results = res_single)
    dev.off()

    #留一法敏感性分析
    pdf(file=".//{path}//pic.leaveoneout.pdf", width=7, height=5.5)
    mr_leaveoneout_plot(leaveoneout_results = mr_leaveoneout(dat))
    dev.off()

        """
    r_script_run = r_script.format(Exposure_id=Exposure_id, Outcome_id=Outcome_id, path=path, gwas_token=gwas_token)

    with open('test.R', 'w', encoding='utf-8') as f:
        f.write(r_script_run)

    os.system('R --slave --no-save --no-restore --no-site-file --no-environ -f  test.R --args')

@timer
def MRtool_MOE(Exposure_id, Outcome_id, path, gwas_token):
    # 等待5s
    time.sleep(5)

    r_script = """
    #引用包
    Sys.setenv(OPENGWAS_JWT="{gwas_token}")
    library(TwoSampleMR)
    library(ieugwasr)
    library(dplyr)
    
    # options(ieugwasr_api = 'gwas-api.mrcieu.ac.uk/')
    
    p_value <- 5e-08
    exposure_dat <- extract_instruments(outcomes = '{Exposure_id}',
                                        p1=p_value,
                                        clump=TRUE,
                                        r2=0.001,
                                        kb=5000,
                                        # access_token = NULL
    )
    
    num_rows <- nrow(exposure_dat)
    print(num_rows)
    if (num_rows < 5) {{
      p_value <- 5e-06
      exposure_dat <- extract_instruments(outcomes = '{Exposure_id}',
                                          p1=p_value,
                                          clump=TRUE,
                                          r2=0.001,
                                          kb=5000,
                                          # access_token = NULL
      )
        num_rows <- nrow(exposure_dat)
        print(num_rows)
    }}
    
    outcomeID="{Outcome_id}"
    #提取结局数据
    outcome_dat <- extract_outcome_data(snps=exposure_dat$SNP, outcomes=outcomeID)
    dat <- harmonise_data(exposure_dat, outcome_dat)
    outTab=dat[dat$mr_keep=="TRUE",]
    write.csv(outTab, file=".//{path}//table.SNP.csv", row.names=F)
    
    
    # Apply all MR methods
    r <- mr_wrapper(dat)
    
    # Load the rf object containing the trained models
    load("rf.rdata")
    # Update the results with mixture of experts
    r <- mr_moe(r, rf)
    
    # Now you can view the estimates, and see that they have
    # been sorted in order from most likely to least likely to
    # be accurate, based on MOE prediction
    r[[1]]$estimates
    head(r)
    
    # save the results csv
    write.csv(r[[1]]$estimates, file=".//{path}//MR.MRresult.csv", row.names=F)
    write.csv(r[[1]]$heterogeneity, file=".//{path}//MR.heterogeneity.csv", row.names=F)
    
    #异质性分析
    heterTab=mr_heterogeneity(dat)
    write.csv(heterTab, file=".//{path}//MR.table.heterogeneity.csv", row.names=F)
    
    #多效性检验
    pleioTab=mr_pleiotropy_test(dat)
    write.csv(pleioTab, file=".//{path}//MR.table.pleiotropy.csv", row.names=F)
    
    # #绘制散点图
    pdf(file=".//{path}//pic.scatter_plot.pdf", width=7.5, height=7)
    mr_scatter_plot(r[[1]]$estimates, dat)
    dev.off()
    
    #森林图
    res_single=mr_singlesnp(dat)      #得到每个工具变量对结局的影响
    pdf(file=".//{path}//pic.forest.pdf", width=7, height=5.5)
    mr_forest_plot(res_single)
    dev.off()
    
    #漏斗图
    pdf(file=".//{path}//pic.funnel_plot.pdf", width=7, height=6.5)
    mr_funnel_plot(singlesnp_results = res_single)
    dev.off()
    
    #留一法敏感性分析
    pdf(file=".//{path}//pic.leaveoneout.pdf", width=7, height=5.5)
    mr_leaveoneout_plot(leaveoneout_results = mr_leaveoneout(dat))
    dev.off()
    """
    r_script_run = r_script.format(Exposure_id=Exposure_id, Outcome_id=Outcome_id, path=path, gwas_token=gwas_token)

    with open('test.R', 'w', encoding='utf-8') as f:
        f.write(r_script_run)

    os.system('R --slave --no-save --no-restore --no-site-file --no-environ -f  test.R --args')

@timer
def MRtool_MRlap(Exposure_id, Outcome_id, path, N_exposure, N_outcome):
    r_script = """
# 安装并加载必要的包
if (!requireNamespace("httr", quietly = TRUE)) {{
  install.packages("httr")
}}
if (!requireNamespace("vcfR", quietly = TRUE)) {{
  install.packages("vcfR")
}}
if (!requireNamespace("MRlap", quietly = TRUE)) {{
  install.packages("MRlap")
}}
if (!requireNamespace("jsonlite", quietly = TRUE)) {{
  install.packages("jsonlite")
}}

library(httr)
library(vcfR)
library(MRlap)
library(jsonlite)

# 定义下载函数
download_vcf <- function(base_url, file_name, dest_dir = ".", min_size = 1 * 1024 * 1024) {{  # min_size 默认 1MB
  # 提取文件标识符 (如 ukb-b-10807)
  file_id <- gsub("(.*)\\\.vcf\\\.gz", "\\\\1", file_name)

  # 构建完整的文件 URL (如 ukb-b-10807/ukb-b-10807.vcf.gz)
  file_url <- paste0(base_url, file_id, "/", file_name)
  print(paste("Attempting to download from:", file_url))

  # 构建文件的本地保存路径
  dest_file <- file.path(dest_dir, file_name)
  print(paste("Saving to:", dest_file))

  # 检查文件是否已经存在并且大小是否合适
  if (file.exists(dest_file)) {{
    file_info <- file.info(dest_file)
    file_size <- file_info$size

    # 如果文件存在且大小大于 1MB，则不重新下载
    if (file_size >= min_size) {{
      message(paste("File already exists and is larger than", round(min_size / (1024 * 1024), 2), "MB. Skipping download."))
      return(TRUE)  # 文件已经存在且大小合适，跳过下载
    }} else {{
      message(paste("File exists but is too small (", round(file_size / (1024 * 1024), 2), "MB). Redownloading..."))
    }}
  }}

  # 尝试直接下载文件，并捕获可能的错误
  tryCatch({{
    # 使用 curl 来下载并显示进度条
    download.file(file_url, destfile = dest_file, method = "curl", mode = "wb")

    # 下载完成后检查文件大小
    file_info <- file.info(dest_file)
    file_size <- file_info$size

    # 如果文件小于 1MB，视为下载失败
    if (file_size < min_size) {{
      message(paste("File size is too small (", round(file_size / (1024 * 1024), 2), "MB). Deleting file:", dest_file))
      file.remove(dest_file)  # 删除文件
      return(FALSE)  # 返回 FALSE 表示下载失败
    }} else {{
      message(paste("Download complete:", dest_file, "File size:", round(file_size / (1024 * 1024), 2), "MB"))
      return(TRUE)  # 返回 TRUE 表示下载成功
    }}

  }}, error = function(e) {{
    # 如果出现错误（如 404），处理错误
    if (grepl("404", e$message)) {{
      message(paste("File not found (404):", file_name))
    }} else {{
      message(paste("Error downloading file:", file_name, "Error message:", e$message))
    }}
    return(FALSE)  # 返回 FALSE 表示下载失败
  }})
}}

# 定义提取数据的函数
extract_data_from_vcf <- function(vcf_file, N) {{
  # 读取 VCF 文件
  vcf <- read.vcfR(vcf_file)

  # 查看样本列名，确保 sample_name 存在
  sample_columns <- colnames(vcf@gt)
  sample_name <- sample_columns[-1]  # 使用第一个样本列
  print("Available sample columns in the VCF:")
  print(sample_name)  # 打印使用的样本列名
  
  # 提取固定字段（CHROM, POS, ID, REF, ALT）
  fix_data <- as.data.frame(vcf@fix)

  # 检查 POS 列是否可以正确转换为数值
  print("Checking POS column:")
  print(summary(fix_data$POS))  # 打印 POS 列的摘要信息
  fix_data$POS <- as.numeric(fix_data$POS)
  if (any(is.na(fix_data$POS))) {{
    warning("Some positions (POS) could not be converted to numeric. Check the VCF file for inconsistencies.")
  }}

  # 提取基因型字段（ES, SE, LP, AF, ID）
  gt_data <- as.data.frame(vcf@gt)

  # 解析 GT 字段中的 ES, SE, LP, AF, ID
  gt_parsed <- do.call(rbind, strsplit(gt_data[, sample_name], ":"))

  # 检查 gt_parsed 的行数是否与 fix_data 的行数一致
  if (nrow(gt_parsed) != nrow(fix_data)) {{
    stop("The number of rows in the genotype data does not match the fixed fields. Check the VCF file for inconsistencies.")
  }}

  # 创建一个 data.frame，符合 MRlap 的要求
  df <- data.frame(
    chr = fix_data$CHROM,               # 染色体编号
    pos = fix_data$POS,                 # 位置
    rsid = fix_data$ID,                 # SNP 标识符
    ref = fix_data$REF,                 # 参考等位基因
    alt = fix_data$ALT,                 # 替代等位基因
    beta = as.numeric(gt_parsed[, 1]),  # 效应量 (ES)
    se = as.numeric(gt_parsed[, 2]),    # 标准误差 (SE)
    zscore = as.numeric(gt_parsed[, 1]) / as.numeric(gt_parsed[, 2]),  # 计算 Z 分数
    N = N                               # 样本量 (TotalControls + TotalCases)
  )

  print(head(df))  # 查看生成的 data.frame
  print("Data extraction complete.")

  return(df)
}}

# 定义基础 URL 和文件名
base_url <- "https://gwas.mrcieu.ac.uk/files/"
exposure_file_name <- "{Exposure_id}.vcf.gz"
outcome_file_name <- "{Outcome_id}.vcf.gz"

# 定义暴露和结果数据的样本量
N_exposure <- {N_exposure}  # 替换为你的暴露数据样本量
N_outcome <- {N_outcome}  # 替换为你的结果数据样本量

# 定义 LD 和 HapMap3 文件路径
ld_file <- "./eur_w_ld_chr"  # 替换为你的 LD 文件路径
hm3_file <- "./w_hm3.snplist"  # 替换为你的 HapMap3 文件路径

# 下载暴露数据文件
if (download_vcf(base_url, exposure_file_name)) {{
  # 下载成功或文件已存在，提取暴露数据
  exposure_data <- extract_data_from_vcf(exposure_file_name, N_exposure)
}} else {{
  stop("Failed to download exposure data.")
}}

# 下载结果数据文件
if (download_vcf(base_url, outcome_file_name)) {{
  # 下载成功或文件已存在，提取结果数据
  outcome_data <- extract_data_from_vcf(outcome_file_name, N_outcome)
}} else {{
  stop("Failed to download outcome data.")
}}

# 使用 MRlap 进行分析
result <- MRlap(
  exposure = exposure_data,
  exposure_name = "Exposure",
  outcome = outcome_data,
  outcome_name = "Outcome",
  ld = ld_file,
  hm3 = hm3_file,
  MR_threshold = 5e-6  # 设置 MR 阈值
)

# 查看结果
print(result)

# 保存结果
# write.csv(result, file = "MRlap_results.csv", row.names = FALSE)
result_json <- toJSON(result, pretty=TRUE)
write(result_json, file = ".//{path}//MRlap_results.json")
    """
    r_script_run = r_script.format(Exposure_id=Exposure_id, Outcome_id=Outcome_id, path=path, N_exposure=N_exposure,
                                   N_outcome=N_outcome)
    # print(r_script_run)

    with open('test.R', 'w', encoding='utf-8') as f:
        f.write(r_script_run)

    os.system('R --slave --no-save --no-restore --no-site-file --no-environ -f  test.R --args')


def get_synonyms(term, api_key):
    try:
        # 获取cui
        url = "https://uts-ws.nlm.nih.gov/rest/search/current?apiKey={apiKey}&string={term}&pageNumber=1&pageSize=1".format(
            apiKey=api_key, term=term)
        payload = {}
        headers = {}
        response = requests.request("GET", url, headers=headers, data=payload)
        # print(response.text.encode('utf8'))
        cui = response.json()["result"]["results"][0]["ui"]
        # print(cui)

        # 获取同义词
        url = "https://uts-ws.nlm.nih.gov/rest/content/current/CUI/{cui}/atoms?apiKey={apiKey}&ttys=&language=ENG&pageSize=25".format(
            apiKey=api_key, cui=cui)
        payload = {}
        headers = {}
        response = requests.request("GET", url, headers=headers, data=payload)
        # print(response.json())
        synonyms = [result["name"] for result in response.json()["result"]]

        # 全体小写
        synonyms = [synonym.lower() for synonym in synonyms]

        # 去重
        synonyms = list(set(synonyms))

    except Exception as e:
        print(e)
        synonyms = []

    # print(len(synonyms))
    print(synonyms)

    return synonyms



from .agent_workflow import MRAgent
import pandas as pd
import os

class MRAgentOE(MRAgent):
    def __init__(self, mode='OE', *args, **kwargs):
        super().__init__(mode, **kwargs)

    def step1(self):
        print(self.path)
        # 新建一个df然后保存到Exposure_and_Outcome.csv
        # 创建一个带有数据的DataFrame
        data = {'index': [1], 'Outcome': [self.outcome], 'Exposure': [self.exposure], 'title': ['title null'],
                'oeID': [1]}
        df = pd.DataFrame(data)
        df.to_csv(os.path.join(self.path, 'Exposure_and_Outcome.csv'), index=False)  # 保存到csv文件

        # 打印DataFrame
        print(df)



MRorNot_text = """Here are articles about {Exposure} and {Outcome} from PubMed, presented in JSON format with the index, title, and abstract of each article. 
Please review each paper sequentially. If the Exposure: {Exposure} and Outcome: {Outcome} have been studied using Mendelian randomization, print the title of the relevant study. If none of the articles have studied the exposure and outcome, print "Exposure and Outcome were not subjected to Mendelian randomization." 
Please note that you only need to output either the title or the phrase "Exposure and Outcome were not subjected to Mendelian randomization.", without any additional words or symbols. 
Once again, please carefully evaluate each paper and output the results accordingly.

json file:
{pubmed_out}"""

pubmed_text = """Below I will give index, title and abstract of most recent {num} paper related to {Outcome} in json format.
Please read each paper in turn, help me find the {Outcome} with what has done correlation studies but not causal inference, which can be further Mendelian randomization studies in the future.

Carefully read each paper and output the judgment results for each article. If a Mendelian randomization study can be performed, explicitly output Outcome and Exposure and the title of the corresponding article. If there are more than one Exposure and Outcome in an article, you should output more than one Exposure and Outcome, as well as the title of the corresponding article. 

The output json list format is as follows:
[{{"index":xxx, "Outcome": xxx, "Exposure": xxx, "title": xxx}},{{"index":xxx, "Outcome": xxx, "Exposure": xxx, "title": xxx}}, …, {{"index":xxx, "Outcome": xxx, "Exposure": xxx, "title": xxx}}]
        
Again, carefully judge each paper and output the results. The results must be output strictly in the required format.

json file:
{pubmed_out}"""

pubmed_text_obo = """Below I will give title and abstract of most recent paper related to {Outcome}.
Please read the paper and help me determine if {Outcome} in the paper has done a correlation study with an influencing factor or another disease, but no causal inference has been made. This can be done in the future with further Mendelian randomisation studies.
Carefully read paper and output the judgment results. If a Mendelian randomization study can be performed, explicitly output Outcome and Exposure.Here Outcome or Exposure should be related to the {Outcome}. If there are more than one Exposure and Outcome in an article, you should output more than one Exposure and Outcome, as well as the title of the corresponding article. 
The output json list format is as follows:
[{{"Outcome": "xxx", "Exposure": "xxx"}},{{"Outcome": "xxx", "Exposure": "xxx"}}, …, {{"Outcome": "xxx", "Exposure": "xxx"}}]
If you think that the paper does not meet the requirements, please output "[{{"Outcome": null, "Exposure": null}}]" directly.

title:
{title}

abstract:
{abstract}

Again, carefully judge each paper and output the results. The results must be output strictly in the required format. Please do not output any extra characters.
"""

synonyms_text = """Please help me find synonyms for "{OE}" in the medical field. If the number of synonyms is more than 7, just output the 7 most relevant synonyms. Output in csv list format.
Note that you only need to output synonyms and commas in csv format, not any other text or symbols(including '' "" . :)."""

gwas_id_text = """When conducting a Mendelian randomization study, it is often necessary to find the SNPs associated with the study in the OpenGWAS database and record the GWAS ID.
Below is the result of searching in OpenGWAS using the keyword {keyword}, output in json list format, saving the GWAS ID,Year,Trait,Consortium,Sample size,Number of SNPs.
I want to get the most suitable GWAS ID, please help me with the following tasks.
1. You are asked to read the Trait and find out if there is a Trait that directly matches {keyword}, and if not, find the Trait that is most relevant to {keyword}.
2. If there are several matching or related Trait at the same time, these can be output at the same time.
3. Output the final corresponding GWAS ID. Output in python list format.

Note that you only need to output gwas_id and the comma in python list format:
['gwas_id1', 'gwas_id2', …, 'gwas_idn']

There is no need to output any other text, including steps, explanations, thought processes, etc. There is no need to output any additional symbols. (including '' "" . :).

json list:
{json_list}"""

LLM_MR_template = """Below is a csv file of the results of Mendelian randomization analysis with Exposure: {Exposure} (GWAS ID: {Exposure_id}) and Outcome: {Outcome} (GWAS ID: {Outcome_id}), please read and understand the results in the table and generate a paragraph describing the results in detail in academic language. Typically, significance can be judged by the Inverse variance weighted method of pval<0.05. or>1 there is a positive correlation and or<1 there is a negative correlation.

#Mendelian randomization analysis results table
{MRresult}

#Heterogeneity analysis results table
{heterogeneity}

#Pleiotropy test results table
{pleiotropy}"""


LLM_MR_MOE_template = """Below is a csv file of the results of Mendelian randomization analysis with Exposure: {Exposure} (GWAS ID: {Exposure_id}) and Outcome: {Outcome} (GWAS ID: {Outcome_id}).
Mendelian randomization analysis was performed based on the Mixture of experts (MOE) function in TwoSampleMR. Once all MR methods have been applied to a summary set, you can then use the mixture of experts to predict the method most likely to be the most accurate. The MOE column, which is a predictor for each method for how well it performs in terms of high power and low type 1 error (scaled 0-1, where 1 is best performance). It orders the table to be in order of best performing method.
Please read and understand the results in the table and generate a paragraph describing the results in detail in academic language.

#Mendelian randomization analysis results table
{MRresult}

#Heterogeneity analysis results table
{heterogeneity}

#Pleiotropy test results table
{pleiotropy}"""

LLM_conclusion_template = """Below is all results report of Mendelian randomization analysis with Exposure: {Exposure} and Outcome: {Outcome}.
I conducted MR studies using different SNPs from Outcome and Exposure in the openGWAS database. Below are the results of the study for the different SNPs, differentiated by GWAS ID. Please help me to summarise the above results and write CONCLUSION in academic language to analyse whether there is a causal effect of Outcome: {Outcome} and Exposure: {Exposure}.
If a causal effect exists, it should be written which Exposure_id and Outcome_id are present in the analysis. If there is no causal effect, it should be made clear that there is no evidence to support a causal inference between Exposure and Outcome.
Try to specify in the analysis which SNPs (Exposure_id and Outcome_id) got this sentence analysed.

# Mendelian randomization analysis results
{MRresult}
"""

LLM_Introduction_template = """I have completed a Mendelian randomization analysis of Exposure: {Exposure} and Outcome: {Outcome}, and I am going to write a paper on this analysis. Please help me write an introduction according to the following:
1. Introduce the current state of research on {Outcome} and its shortcomings
2. Propose the hypothesis that Outcome: {Outcome} may be related to Exposure: {Exposure}.
3. Analyse the effect of Exposure: {Exposure} on Outcome: {Outcome} using Mendelian Randomization analysis.
Note that you should refer to the literature given below when writing the Introduction and label the referenced literature in APA format.

Below I found some related articles on PubMed for your reference:
About {Exposure}.
{Exposure_pubmed}

About {Outcome}.
{Outcome_pubmed}

About {Exposure} and {Outcome}.
{Exposure_Outcome_pubmed}
"""

LLM_template_MR_effect_evaluation = """
Below is an article on a Mendelian randomization study. Please read it carefully and assess its adherence to the STROBE-MR checklist. For each item on the checklist, return a JSON object where each item (including sub-items, such as 4(a)) is marked as "yes" or "no" based on the article's content. Only provide the JSON output without any additional explanation or commentary. This means you should strictly return the JSON object without any text before or after it.

STROBE-MR Checklist Items:

TITLE and ABSTRACT:
   - 1. Indicate Mendelian randomization (MR) as the study’s design in the title and/or the abstract if that is a main purpose of the study.

INTRODUCTION:
   - 2. Background: Explain the scientific background and rationale for the reported study. What is the exposure? Is a potential causal relationship between exposure and outcome plausible? Justify why MR is a helpful method to address the study question.
   - 3. Objectives: State specific objectives clearly, including pre-specified causal hypotheses (if any). State that MR is a method that, under specific assumptions, intends to estimate causal effects.

METHODS:
   - 4. Study design and data sources:
     a) Setting: Describe the study design and the underlying population, if possible. Describe the setting, locations, and relevant dates, including periods of recruitment, exposure, follow-up, and data collection, when available.
     b) Participants: Give the eligibility criteria, and the sources and methods of selection of participants. Report the sample size, and whether any power or sample size calculations were carried out prior to the main analysis.
     c) Describe measurement, quality control and selection of genetic variants.
     d) For each exposure, outcome, and other relevant variables, describe methods of assessment and diagnostic criteria for diseases.
     e) Provide details of ethics committee approval and participant informed consent, if relevant.

   - 5. Assumptions: Explicitly state the three core IV assumptions for the main analysis (relevance, independence and exclusion restriction) as well assumptions for any additional or sensitivity analysis.

   - 6. Statistical methods: main analysis:
     a) Describe how quantitative variables were handled in the analyses (i.e., scale, units, model).
     b) Describe how genetic variants were handled in the analyses and, if applicable, how their weights were selected.
     c) Describe the MR estimator (e.g., two-stage least squares, Wald ratio) and related statistics. Detail the included covariates and, in case of two-sample MR, whether the same covariate set was used for adjustment in the two samples.
     d) Explain how missing data were addressed.
     e) If applicable, indicate how multiple testing was addressed.

   - 7. Assessment of assumptions: Describe any methods or prior knowledge used to assess the assumptions or justify their validity.

   - 8. Sensitivity analyses and additional analyses: Describe any sensitivity analyses or additional analyses performed (e.g., comparison of effect estimates from different approaches, independent replication, bias analytic techniques, validation of instruments, simulations).

   - 9. Software and pre-registration:
     a) Name statistical software and package(s), including version and settings used.
     b) State whether the study protocol and details were pre-registered (as well as when and where).

RESULTS:
   - 10. Descriptive data:
     a) Report the numbers of individuals at each stage of included studies and reasons for exclusion. Consider use of a flow diagram.
     b) Report summary statistics for phenotypic exposure(s), outcome(s), and other relevant variables (e.g., means, SDs, proportions).
     c) If the data sources include meta-analyses of previous studies, provide the assessments of heterogeneity across these studies.
     d) For two-sample MR:
        i. Provide justification of the similarity of the genetic variant-exposure associations between the exposure and outcome samples.
        ii. Provide information on the number of individuals who overlap between the exposure and outcome studies.

   - 11. Main results:
     a) Report the associations between genetic variant and exposure, and between genetic variant and outcome, preferably on an interpretable scale.
     b) Report MR estimates of the relationship between exposure and outcome, and the measures of uncertainty from the MR analysis, on an interpretable scale, such as odds ratio or relative risk per SD difference.
     c) If relevant, consider translating estimates of relative risk into absolute risk for a meaningful time period.
     d) Consider plots to visualize results (e.g., forest plot, scatterplot of associations between genetic variants and outcome versus between genetic variants and exposure).

   - 12. Assessment of assumptions:
     a) Report the assessment of the validity of the assumptions.
     b) Report any additional statistics (e.g., assessments of heterogeneity across genetic variants, such as I2, Q statistic or E-value).

   - 13. Sensitivity analyses and additional analyses:
     a) Report any sensitivity analyses to assess the robustness of the main results to violations of the assumptions.
     b) Report results from other sensitivity analyses or additional analyses.
     c) Report any assessment of direction of causal relationship (e.g., bidirectional MR).
     d) When relevant, report and compare with estimates from non-MR analyses.
     e) Consider additional plots to visualize results (e.g., leave-one-out analyses).

DISCUSSION:
   - 14. Key results: Summarize key results with reference to study objectives.
   - 15. Limitations: Discuss limitations of the study, taking into account the validity of the IV assumptions, other sources of potential bias, and imprecision. Discuss both direction and magnitude of any potential bias and any efforts to address them.
   - 16. Interpretation:
     a) Meaning: Give a cautious overall interpretation of results in the context of their limitations and in comparison with other studies.
     b) Mechanism: Discuss underlying biological mechanisms that could drive a potential causal relationship between the investigated exposure and the outcome, and whether the gene-environment equivalence assumption is reasonable. Use causal language carefully, clarifying that IV estimates may provide causal effects only under certain assumptions.
     c) Clinical relevance: Discuss whether the results have clinical or public policy relevance, and to what extent they inform effect sizes of possible interventions.

   - 17. Generalizability: Discuss the generalizability of the study results (a) to other populations, (b) across other exposure periods/timings, and (c) across other levels of exposure.

OTHER INFORMATION:
   - 18. Funding: Describe sources of funding and the role of funders in the present study and, if applicable, sources of funding for the databases and original study or studies on which the present study is based.
   - 19. Data and data sharing: Provide the data used to perform all analyses or report where and how the data can be accessed, and reference these sources in the article. Provide the statistical code needed to reproduce the results in the article, or report whether the code is publicly accessible and if so, where.
   - 20. Conflicts of Interest: All authors should declare all potential conflicts of interest.

Example Output: {{"1": "yes", "2": "no", "3": "yes", "4a": "yes", "4b": "no", "4c": "yes", "4d": "yes", "4e": "no", "5": "yes", "6a": "no", "6b": "yes", "6c": "yes", "6d": "no", "6e": "yes", "7": "no", "8": "yes", "9a": "yes", "9b": "no", "10a": "yes", "10b": "no", "10c": "yes", "10d_i": "no", "10d_ii": "yes", "11a": "yes", "11b": "no", "11c": "yes", "11d": "no", "12a": "yes", "12b": "no", "13a": "yes", "13b": "no", "13c": "yes", "13d": "no", "13e": "yes", "14": "no", "15": "yes", "16a": "no", "16b": "yes", "16c": "no", "17": "yes", "18": "no", "19": "yes", "20": "no"}}


The following is the full text, recorded in json format:
{paper_details}
"""




from openai import OpenAI
import google.generativeai as genai
import anthropic


# 调用GPT
def openAI_gpt(text, openAI_key, model_name='gpt-4-1106-preview', base_url=None):
    openai_api_key = openAI_key
    if base_url is not None:
        client = OpenAI(api_key=openai_api_key, base_url=base_url)
    else:
        client = OpenAI(api_key=openai_api_key)

    chat_response = client.chat.completions.create(
        model=model_name,
        seed=42,
        messages=[
            {"role": "system", "content": "You are a helpful biomedical scientist."},
            {"role": "user", "content": text}, ]
    )

    return chat_response.choices[0].message.content


# 调用Ollama
def ollama_chat(text, model_name='llama2'):
    import ollama
    response = ollama.chat(
        model=model_name,
        messages=[{"role": "system", "content": "You are a helpful biomedical scientist."},
                  {"role": "user", "content": text}, ]
    )
    return response['message']['content']


# 调用Gemini
def gemini_chat(text, model_name='gemini-1.5-pro-latest', AI_key=None):
    # sleep 1s to avoid 429 error
    # time.sleep(10)
    genai.configure(api_key=AI_key)
    model = genai.GenerativeModel(model_name)
    response = model.generate_content(text)
    return response.text


# 调用Anthropic
def anthropic_chat(text, model_name='claude-3-opus-20240229', AI_key=None):
    client = anthropic.Anthropic(api_key=AI_key, )
    message = client.messages.create(
        model=model_name,
        max_tokens=1000,
        messages=[{"role": "system", "content": "You are a helpful biomedical scientist."},
                  {"role": "user", "content": text}]
    )
    return message.content


def anthropic_chat_openai(text, AI_key, model_name='claude-3-opus-20240229'):
    client = OpenAI(api_key=AI_key, base_url="https://api.ai365vip.com/v1/")
    chat_response = client.chat.completions.create(
        model=model_name,
        seed=42,
        messages=[{"role": "system", "content": "You are a helpful biomedical scientist."},
                  {"role": "user", "content": text}]
    )
    print(chat_response)

    return chat_response.choices[0].message.content



def qwen_chat(text, AI_key, model_name='qwen-max-0403'):
    openai_api_key = AI_key
    client = OpenAI(api_key=openai_api_key, base_url="https://dashscope.aliyuncs.com/compatible-mode/v1")

    chat_response = client.chat.completions.create(
        model=model_name,
        seed=42,
        messages=[
            {"role": "system", "content": "You are a helpful biomedical scientist."},
            {"role": "user", "content": text}, ]
    )

    return chat_response.choices[0].message.content

def llm_chat(text, model_name, AI_key=None, base_url=None):
    if 'gpt' in model_name:
        return openAI_gpt(text, AI_key, model_name, base_url)
    elif 'gemini' in model_name:
        return gemini_chat(text, model_name, AI_key)
    elif 'claude' in model_name:
        return anthropic_chat(text, model_name, AI_key)
    elif 'qwen-max' in model_name:
        return qwen_chat(text, AI_key, model_name)
    else:
        return ollama_chat(text, model_name)
